#include "hourglass.h"

static void usage (void)
{
  printf ("usage: hourglass <options>\n");
  printf ("  -n <number of threads to create>\n");
  printf ("  -d <duration of experiment>\n");
  printf ("  -t <thread number for subsequent options to affect>\n");
  printf ("  -a -- subsequent options to affect all threads\n");
  printf ("    -p <priority>\n");
  printf ("      IDLE, LOW, NORMAL, HIGH, HIGHEST (timesharing priorities)\n");
  printf ("      RTLOW, RTMID, RTHIGH (real-time priorities)\n");
  printf ("    -w <workload>\n");
  printf ("      LAT -- scheduling latency test\n");
  printf ("      CPU -- cpu-bound execution trace\n");
  printf ("      PERIODIC <period> <amount> -- periodic task, execution trace\n");
  printf ("      CPU_PERIODIC <gap> <amount> -- cpu-bound periodic\n");
  printf ("      YIELD <amount> -- yield CPU after receiving amount, execution trace\n");
  printf ("    -r <period> <amount> -- request a CPU reservation\n");
  printf ("    -l -- log thread events, if appropriate\n");
  printf ("    -i <kind of timers to use>\n");
  printf ("      HIGHRES -- precise timers, multimedia timers, or equivalent\n");
  printf ("      OS -- timed wait based on OS timer interrupt\n");
  printf ("      BUILTIN -- high-res timer supplied by hourglass\n");
  // printf ("    -i <kind of timer> -- over-rides global timer preference\n");
  printf ("\n");
  printf ("Times must be of the form 1.5s, 1020ms, 87.0us.\n");
  printf ("\n");
  printf ("See MANUAL for more detail.\n");
  exit (-1);
}

static void init_thrd_data (void)
{
  int i;

  thrd_data = (struct thrd_data_t *) 
    xmalloc (num_threads * sizeof (struct thrd_data_t));

  for (i=0; i<num_threads; i++) {
    // initialize vars
    thrd_data[i].lat_slot = 0;
    thrd_data[i].exec_start = 0;
    thrd_data[i].last_exec = 0;
    thrd_data[i].reserved_record = -1;
    thrd_data[i].lat_slots = NULL;

    // set defaults
    thrd_data[i].log = 0;
    thrd_data[i].pri = NORMAL;
    thrd_data[i].task = CPU;
  }
}

static int expect_int (int argc, char *argv[], int pos)
{
  int i;
  char *c;

  if (pos >= argc) {
      printf ("oops: integer expected at cmd line position %d\n",
	      pos);
      exit (-1);
  }

  i = strtol (argv[pos], &c, 10);
  if (*c != '\0') {
    printf ("oops: integer expected at cmd line position %d\n",
	    pos);
    exit (-1);
  }

  return i;
}

static us_time expect_time (int argc, char *argv[], int pos)
{
  int len;
  double t;
  char *c;

  if (pos >= argc) {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
  }

  len = strlen (argv[pos]);

  // microseconds
  if (len >= 3 &&
      strcmp (&argv[pos][len-2], "us") == 0) {
    argv[pos][len-2] = '\0';
    t = strtod (argv[pos], &c);
    if (*c != '\0') {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
    }
    return (us_time) t;
  }

  // milliseconds
  if (len >= 3 &&
      strcmp (&argv[pos][len-2], "ms") == 0) {
    argv[pos][len-2] = '\0';
    t = strtod (argv[pos], &c);
    if (*c != '\0') {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
    }
    return (us_time) t * 1000;
  }

  // seconds - this needs to be last
  if (len >= 2 &&
      strcmp (&argv[pos][len-1], "s") == 0) {
    argv[pos][len-1] = '\0';
    t = strtod (argv[pos], &c);
    if (*c != '\0') {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
    }
    return (us_time) t * 1000000;
  }

  printf ("oops: time value expected at cmd line position %d\n",
	  pos);
  exit (-1);
}

static enum thrd_pri_t expect_pri (int argc, char *argv[], int pos)
{
}

void parse_args (int argc, char *argv[])
{
  int i, pos;
  int cur_thread = -1;

  if ((argc < 2) ||
      (strcmp (argv[1], "-h") == 0) ||
      (strcmp (argv[1], "-?") == 0)) {
    usage ();
  }

  pos = 1;
  while (pos < argc) {

    if (strcmp (argv[pos], "-n") == 0) {
      pos++;
      num_threads = expect_int (argc, argv, pos);
      if (num_threads > 50) {
	printf ("oops: max of 50 threads: too many have been observed\n");
	printf ("to lock up the Linux kernel.\n");
	exit (-1);
      }
      pos++;
      printf ("%d threads\n", num_threads);
      init_thrd_data ();
      continue;
    }

    if (strcmp (argv[pos], "-d") == 0) {
      pos++;
      duration = (unsigned int) expect_time (argc, argv, pos);
      pos++;
      printf ("%d us\n", duration);
      continue;
    }

    if (strcmp (argv[pos], "-t") == 0) {
      pos++;
      if (thrd_data) {
	printf ("oops: option -t may appear once, and after -n\n");
	exit (-1);
      }
      cur_thread = expect_int (argc, argv, pos);
      pos++;
      if (cur_thread < 0 || cur_thread >= num_threads) {
	printf ("oops: thread number %d out of range\n", cur_thread);
	exit (-1);
      }
      continue;
    }

    if (strcmp (argv[pos], "-p") == 0) {
      if (cur_thread == -1) {
	printf ("oops: option %d (-p) is only meaningful for a thread\n",
		pos);
	exit (-1);
      }
      pos++;
      thrd_data[cur_thread].pri = expect_pri (argc, argv, pos);
      pos++;
      continue;
    }

    printf ("oops: unexpected arg at position %d\n", pos);
    exit (-1);
  }

#if 0
  if (strcmp (argv[1], "lat") == 0) {

    init_thrd_data (1);
    thrd_data[0].task = LAT;
    thrd_data[0].pri = HIGHEST;
    thrd_data[0].lat_slots = (cycle_time *)
      xmalloc (sizeof (cycle_time) * LAT_SLOTS);

  } else if (strcmp (argv[1], "per1") == 0) {    

    init_thrd_data (1);
    thrd_data[0].task = PERIODIC;
    thrd_data[0].pri = HIGHEST;

  } else if (strcmp (argv[1], "per10") == 0) {    

    init_thrd_data (10);
    for (i=0; i<10; i++) {
      thrd_data[i].task = PERIODIC;
      thrd_data[i].pri = HIGHEST;
    }

  } else if (strcmp (argv[1], "cpu1") == 0) {    

    init_thrd_data (1);
    thrd_data[0].task = CPU;
    need_timers = 0;

  } else if (strcmp (argv[1], "cpu10") == 0) {    

    init_thrd_data (10);
    for (i=0; i<10; i++) {
      thrd_data[i].task = CPU;
      thrd_data[i].pri = HIGHEST;
    }
    need_timers = 0;

  } else if (strcmp (argv[1], "yield10") == 0) {    

    init_thrd_data (10);
    for (i=0; i<10; i++) {
      thrd_data[i].task = YIELD;
      // thrd_data[i].pri = HIGHEST;
    }
    need_timers = 0;

  } else {
    usage ();
  }
#endif

  tracing = 0;
  for (i=0; i<num_threads; i++) {
    if (thrd_data[i].task == YIELD ||
	thrd_data[i].task == CPU ||
	thrd_data[i].task == PERIODIC) {
      tracing = 1;
    }
  }

  if (tracing) {
    trace_recs = (struct trace_rec_t *)
      xmalloc (sizeof (struct trace_rec_t) * NUM_TRACE_RECORDS);

#ifdef GAP_HIST
    gap_hist = (int *) 
      xmalloc (sizeof (int) * NUM_GAP_HIST_BUCKETS);    
    for (i=0; i<NUM_GAP_HIST_BUCKETS; i++) {
      gap_hist[i] = 0;
    }
#endif

  } else {
    trace_recs = NULL;
#ifdef GAP_HIST
    gap_hist = NULL;
#endif
  }
  current_record = 0;
}

#if 0
static void old_parse_args (int argc, char *argv[])
{
  int i;

  if (argc != 2) {
    usage ();
  }

  if (strcmp (argv[1], "lat") == 0) {

    duration = 10000000;
    init_thrd_data (1);
    thrd_data[0].task = LAT;
    thrd_data[0].pri = HIGHEST;
    thrd_data[0].lat_slots = (cycle_time *)
      xmalloc (sizeof (cycle_time) * LAT_SLOTS);

  } else if (strcmp (argv[1], "per1") == 0) {    

    duration = 10000000;
    init_thrd_data (1);
    thrd_data[0].task = PERIODIC;
    thrd_data[0].pri = HIGHEST;

  } else if (strcmp (argv[1], "per10") == 0) {    

    duration = 10000000;
    init_thrd_data (10);
    for (i=0; i<10; i++) {
      thrd_data[i].task = PERIODIC;
      thrd_data[i].pri = HIGHEST;
    }

  } else if (strcmp (argv[1], "cpu1") == 0) {    

    duration = 10000000;
    init_thrd_data (1);
    thrd_data[0].task = CPU;
    need_timers = 0;

  } else if (strcmp (argv[1], "cpu10") == 0) {    

    duration = 10000000;
    init_thrd_data (10);
    for (i=0; i<10; i++) {
      thrd_data[i].task = CPU;
      thrd_data[i].pri = HIGHEST;
    }
    need_timers = 0;

  } else if (strcmp (argv[1], "yield10") == 0) {    

    duration = 10000000;
    init_thrd_data (10);
    for (i=0; i<10; i++) {
      thrd_data[i].task = YIELD;
      // thrd_data[i].pri = HIGHEST;
    }
    need_timers = 0;

  } else {
    usage ();
  }

  tracing = 0;
  for (i=0; i<num_threads; i++) {
    if (thrd_data[i].task == YIELD ||
	thrd_data[i].task == CPU ||
	thrd_data[i].task == PERIODIC) {
      tracing = 1;
    }
  }

  if (tracing) {
    trace_recs = (struct trace_rec_t *)
      xmalloc (sizeof (struct trace_rec_t) * NUM_TRACE_RECORDS);

#ifdef GAP_HIST
    gap_hist = (int *) 
      xmalloc (sizeof (int) * NUM_GAP_HIST_BUCKETS);    
    for (i=0; i<NUM_GAP_HIST_BUCKETS; i++) {
      gap_hist[i] = 0;
    }
#endif

  } else {
    trace_recs = NULL;
#ifdef GAP_HIST
    gap_hist = NULL;
#endif
  }
  current_record = 0;
}
#endif
